package aexp.steps;

import android.app.Service;
import android.content.Intent;
import android.content.Context;
import android.content.SharedPreferences;
import android.content.IntentFilter;
import android.content.BroadcastReceiver;
import android.hardware.Sensor;
import android.hardware.SensorEvent;
import android.hardware.SensorEventListener;
import android.os.PowerManager;
import android.os.IBinder;
import android.os.RemoteException;
import android.os.DeadObjectException;
import android.hardware.SensorManager;
import android.util.Log;
import android.util.LogPrinter;
import java.io.PrintWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.List;
import java.util.Date;

public class SamplingService extends Service implements SensorEventListener {
	static final String LOG_TAG = "STEPS_SAMPLINGSERVICE";
	static final boolean DEBUG_GENERAL = true;

	public int onStartCommand(Intent intent, int flags, int startId) {
		super.onStartCommand( intent, flags, startId );
		Log.d( LOG_TAG, "onStartCommand" );
		stopSampling();		// just in case the activity-level service management fails
        rate = SensorManager.SENSOR_DELAY_FASTEST;
		sensorManager = (SensorManager)getSystemService( SENSOR_SERVICE  );
		startSampling();
		Log.d( LOG_TAG, "onStartCommand ends" );
		return START_NOT_STICKY;
	}

	public void onDestroy() {
		super.onDestroy();
		Log.d( LOG_TAG, "onDestroy" );
		stopSampling();
	}

	public IBinder onBind(Intent intent) {
		return serviceBinder;	// cannot bind
	}

// SensorEventListener
    public void onAccuracyChanged (Sensor sensor, int accuracy) {
    }

    public void onSensorChanged(SensorEvent sensorEvent) {
		processSample( sensorEvent.values );
    }


	private void stopSampling() {
		if( !samplingStarted )
			return;
        if( sensorManager != null ) {
			Log.d( LOG_TAG, "unregisterListener/SamplingService" );
            sensorManager.unregisterListener( this );
		}
        if( captureFile != null ) {
            captureFile.close();
			captureFile = null;
        }
		samplingStarted = false;
	}

	private void startSampling() {
		if( samplingStarted )
			return;

        initSampling();
       	List<Sensor> sensors = sensorManager.getSensorList( Sensor.TYPE_ACCELEROMETER  );
      	ourSensor = sensors.size() == 0 ? null : sensors.get( 0 );

      	if( ourSensor != null ) {
			Log.d( LOG_TAG, "registerListener/SamplingService" );
           	sensorManager.registerListener( 
                            this, 
                            ourSensor,
                            rate );
		}
		captureFile = null;
        if( DEBUG_GENERAL ) {
       	    File captureFileName = new File( "/sdcard", "capture.csv" );
       	    try {
         	    captureFile = new PrintWriter( new FileWriter( captureFileName, false ) );
            } catch( IOException ex ) {
       		    Log.e( LOG_TAG, ex.getMessage(), ex );
       	    }
        }
		samplingStarted = true;
	}

    private double x_max;
    private double x_min;
	private long sampleCounter;
    private boolean prev_ref;
    private boolean prev_div;
    private boolean state_ref;
    private boolean state_div;
    private boolean prev_y_d;
    private double conv_min;
    private double conv_max;
    private double conv_power;
    private double conv_power_cycle;
    private int cycle_ctr;
	private double osc_phase;
	private double osc_phase_shift;
    private CircularBuffer mampl;
    private CircularBuffer osc_out_signal;
    private CircularBuffer err_out;
    private SpikeFilter x_d_spike_filter;
    private int steps = 0;
    private double sensitivity = 3.0;

    private void initSampling() {
        x_max = 11.0;
        x_min = 9.0;
        prev_ref = false;
        prev_div = false;
        state_ref = false;
        state_div = false;
        prev_y_d = false;
        conv_min = 0.0;
        conv_max = 0.0;
        conv_power = 0.0;
        conv_power_cycle = 0;
        cycle_ctr = 0;
        osc_phase = 0.0;
        osc_phase_shift = 0.0;
        mampl = new CircularBuffer( 10 );
        osc_out_signal = new CircularBuffer( 10 );
        err_out = new CircularBuffer( 6 );
        x_d_spike_filter = new SpikeFilter( 6 );
        steps = 0;
    }

    private void processAmpl( double ampl ) {
        double mampl_value = ampl - 10.0;
        mampl.putValue( mampl_value );
        double x_mean = updateXmaxMin( ampl );
        boolean x_d = digiSign( ampl,x_mean );
        x_d = x_d_spike_filter.putValue( x_d );
        double y = oscOut();
        boolean y_d = digiSign( y,0.0 );
        osc_out_signal.putValue( y );
        double conv_s = convolution( mampl.getBuffer(), 
                                    mampl.getWritePtr(),
                                    osc_out_signal.getBuffer(),
                                    osc_out_signal.getWritePtr() );
        conv_power += conv_s*conv_s;
        ++cycle_ctr;
        boolean stepDetected = false;
        if( y_d && !prev_y_d ) {
            double max_ampl = x_max - x_min;
            if( max_ampl < 20.0 )
                max_ampl = 20.0;
            conv_power_cycle = ( Math.sqrt( conv_power / (double)cycle_ctr ) / max_ampl ) * 10.0;
            conv_power = 0.0;
            cycle_ctr = 0;
            stepDetected = conv_power_cycle > sensitivity;
        }
        prev_y_d = y_d;
        boolean err_d = phaseDetector( x_d , y_d );
        if( err_d )
            err_out.putValue( 1.0 );
        else
            err_out.putValue( 0.0 );
        double filtered_err = convolution( err_out.getBuffer(),
                                            err_out.getWritePtr(),
                                            lowpassFilter,
                                            0 );
        double phase_tune = (24.0/30.0)*filtered_err;
        osc_phase_shift = Math.PI/30.0+phase_tune;
        if( stepDetected ) {
            ++steps;
            step( steps );
        }
        if( DEBUG_GENERAL ) {
            if( captureFile != null )
                captureFile.println( mampl_value+","+
                                 x_d+","+
                                 y+","+
                                 y_d+","+
                                 conv_power_cycle+","+
                                 err_d+","+
                                 filtered_err+","+
                                 stepDetected+","+
                                 steps );
        }
    }

// Processes one sample
	private void processSample( float values[] ) {
		if( values.length < 3 )
			return;
		double ampl = Math.sqrt( (double)values[0]*values[0] +
								 (double)values[1]*values[1] +
								 (double)values[2]*values[2] );
        processAmpl( ampl );
	}

	private double convolution( double inp[], int inpPtr, double filter[], int filterPtr ) {
        int inpCtr = inpPtr - 1;
        int filterCtr = filterPtr;
		double result = 0.0;
		for( int i = 0 ; i < filter.length ; ++i ) {
			if( inpCtr < 0 )
				inpCtr = inp.length - 1;
            if( filterCtr >= filter.length )
                filterCtr = 0;
			result += filter[filterCtr] * inp[inpCtr];
            --inpCtr;
            ++filterCtr;
		}
		return result;
	}

    private double updateXmaxMin( double ampl ) {
        if( ampl > x_max )
            x_max = ampl;
        else
            x_max = 0.95*x_max+0.05*ampl;
        if( ampl < x_min )
            x_min = ampl;
        else
            x_min = 0.95*x_min+0.05*ampl;
        return ( x_max + x_min ) * 1.05 / 2.0;
    }

    private boolean digiSign( double ampl, double x_mean ) {
        return ampl > x_mean;
    }

    private double oscOut() {
	    double out = Math.cos( osc_phase );
	    osc_phase = osc_phase + osc_phase_shift;
	    return out;
    }

    private boolean phaseDetector( boolean ref, boolean div) {
        if( !prev_ref && ref ) {
            state_ref=true;
            state_div = false;
        }
        if( !prev_div && div )
            state_div=true;
        prev_ref=ref;
        prev_div=div;
        if( state_ref && state_div ) {
            state_ref = false;
            state_div = false;
        }
        return state_ref;
    } 

	private void step( int count ) {
		if( iSteps == null )
			Log.d( LOG_TAG, "step() calback: cannot call back (count: "+count+")" );
		else {
			Log.d( LOG_TAG, "step() callback: count: "+count );
		  	try {
				iSteps.step( count );
		  	} catch( DeadObjectException ex ) {
				Log.e( LOG_TAG,"step() callback", ex );
		  	} catch( RemoteException ex ) {
				Log.e( LOG_TAG, "RemoteException",ex );
			}
		}
	}

    private ISteps iSteps = null;
    private final ISamplingService.Stub serviceBinder = 
			new ISamplingService.Stub() {
		public void setCallback( IBinder binder ) {
			Log.d( LOG_TAG,"setCallback" );
			iSteps = ISteps.Stub.asInterface( binder );
      	}

		public void removeCallback() {
			Log.d( LOG_TAG, "removeCallback" );
			iSteps = null;
		}

		public boolean isSampling() {
			return samplingStarted;
		}

		public void stopSampling() {
			Log.d( LOG_TAG, "stopSampling" );
			SamplingService.this.stopSampling();
			stopSelf();
		}

        public void setSensitivity( double sensitivity ) {
            Log.d( LOG_TAG, "setSensitivity: "+sensitivity );
            SamplingService.this.sensitivity = sensitivity;
        }

        public void setNoiseSensitivity( int noiseSensitivity ) {
            Log.d( LOG_TAG, "setNoiseSensitivity: "+noiseSensitivity );
            SamplingService.this.x_d_spike_filter = new SpikeFilter( noiseSensitivity );
        }

    };


    private int rate;
    private SensorManager sensorManager;
    private PrintWriter captureFile;
	private Sensor ourSensor;
	private boolean samplingStarted = false;

    private double lowpassFilter[] = { 
                    0.02466075,
                    0.09563908,
                    0.23767242,
                    0.30871625,
                    0.23767242,
                    0.09563908 };


    class CircularBuffer {
        public CircularBuffer( int size ) {
            buffer = new double[ size ];
            for( int i = 0 ; i < size ; ++i )
                buffer[i] = 0.0;
        }

        public void putValue( double value ) {
            buffer[ writePtr++ ] = value;
            if( writePtr >= buffer.length )
                writePtr = 0;
        }

        public double[] getBuffer() {
            return buffer;
        }

        public int getWritePtr() {
            return writePtr;
        }

        public int writePtr = 0;
        public double buffer[];
    }


    class SpikeFilter {
        public SpikeFilter( int size ) {
            buffer = new boolean[ size ];
            for( int i = 0 ; i < size ; ++i )
                buffer[i] = false; 
        }

        public boolean putValue( boolean value ) {
            boolean retValue = false;
            int buflen = buffer.length;
            for( int i = 0 ; i < buflen ; ++i )
                if( buffer[i] )
                    retValue = true;
            buffer[writePtr] = value;
            writePtr = ( ++writePtr ) % buflen;
            return retValue;
        }

        private int writePtr = 0;
        private boolean buffer[];
    }

}

